use serialport;
use serialport::prelude::*;

use utilities::prelude::*;

use std::cell::RefCell;
use std::io;

pub use serialport::{DataBits, FlowControl, Parity, SerialPortSettings, StopBits};
pub use std::time::Duration;

pub enum SerialReadResult {
    Message(String),
    Timeout,
}

pub struct Port {
    serial_port: RefCell<Box<dyn SerialPort>>,
}

impl Port {
    pub fn open(settings: SerialPortSettings) -> VerboseResult<Self> {
        let port_path = Self::find_macroboard()?;

        let port = serialport::open_with_settings(&port_path, &settings).map_err(|err| {
            format!(
                "could not open serial port ({:?}, {}, {})",
                err.kind(),
                err,
                port_path
            )
        })?;

        Ok(Port {
            serial_port: RefCell::new(port),
        })
    }

    #[allow(unused)]
    pub fn read(&self) -> VerboseResult<SerialReadResult> {
        let mut buf: Vec<u8> = (0..255).collect();

        match self.serial_port.try_borrow_mut()?.read(&mut buf[..]) {
            Ok(t) => Ok(SerialReadResult::Message(
                String::from_utf8(buf[0..t].to_vec())
                    .map_err(|err| format!("failed converting utf8 buffer ({})", err))?,
            )),
            Err(ref e) if e.kind() == io::ErrorKind::TimedOut => Ok(SerialReadResult::Timeout),
            Err(err) => create_error!(format!("failed reading serial port ({})", err)),
        }
    }

    #[allow(unused)]
    pub fn write(&self, msg: &str) -> VerboseResult<()> {
        self.serial_port
            .try_borrow_mut()?
            .write(msg.as_bytes())
            .map_err(|err| format!("failed writing to serial port ({})", err))?;

        Ok(())
    }

    fn find_macroboard() -> VerboseResult<String> {
        let available_ports = serialport::available_ports()
            .map_err(|err| format!("error querying serial ports ( {})", err))?;

        for available_port in available_ports.iter() {
            if let serialport::SerialPortType::UsbPort(usb_info) = &available_port.port_type {
                // check for the correct device
                if let Some(product_name) = &usb_info.product {
                    if product_name == "FT232R_USB_UART"
                        && usb_info.vid == 1027
                        && usb_info.pid == 24577
                    {
                        return Ok(available_port.port_name.clone());
                    }
                }
            }
        }

        create_error!("macro board not found on usb bus")
    }
}
